import string,re
import clump.db,ldap
from registry import objtype,getObjType
from defaults import getDefaults
from clump.logfile import log

FieldError="Identically named fields found for a CLuMP object"

def parseDN(line,create=0):
    # Parse the commnd line for a DN string
    m=re.compile("^([^=\s]+)(\s+|$)").match(line)
    # If we find one then convert it to a real DN
    if m:
        line=line[len(m.group(0)):]
        dn=clump.db.db.string2dn(m.group(1),create)
        # If there is no DN string then we must be using the current object
    else:
        dn=clump.db.db.currentDN
    return dn

class Field:
    def __init__(self,ldapname,name,desc,create=0,must=0,show=1):
        if create:
            must=1
        self.ldapname=ldapname
        self.name=string.lower(name)
        self.desc=desc
        self.unique=len(self.name)
        self.create=create  # Is this field needed for object creation
        self.must=must      # Is this field required by LDAP
        self.show=show      # Is this field shown?

    def setUnique(self,list):
        i=1
        for field in list:
            if field==self:  # Skip this field if it is found
                continue
            if self.name==field.name:
                raise FieldError
            while self.name[:i]==field.name[:i]:
                i=i+1
        self.unique=i

    def help(self):
        print "    %-10s : %s" % (string.upper(self.name[:self.unique])+
                                  self.name[self.unique:],self.desc)

    def match(self,key):
        if len(key)<self.unique or string.lower(key)!=self.name[:len(key)] \
           or len(key)>len(self.name):
            return 0
        else:
            return -1

class BaseObject:
    def __init__(self,name,objtype,namefield):
        self.name=string.lower(name)
        self.objtype=objtype
        self.namefield=namefield
        try:
            for f in self.fields:
                f.setUnique(self.fields)
        except FieldError:
            print "** ERROR ** Field '%s' defined twice" % f.name

    def parseDefaults(self):
        defs=getDefaults(self.name)
        for key in defs.keys():
            field=self.getLDAPname(key)
            if not field:
                print "** UNKNOWN FIELD ** Ignoring default for '%s' in %s" % (
                    key,self.name)
                continue
            if field!=key:
                defs[field]=defs[key]
                del defs[key]
        return

    def getLDAPname(self,key):
        for f in self.fields:
            if f.match(key):
                return f.ldapname
        return ""

    def parseLine(self,line,create=0,absolute=0):
        if create:
            absolute=1
        dn=parseDN(line,create)
        if not dn:
            return (None,None)
        dict={}
        # Use a regular expression to identify all the keyword=value pairs
        # in the command line and then loop over them
        for attr in re.compile(
            "([^=+\s-]+)\s*(=|\+=|-=)\s*((\"[^\"]*\")|(\'[^\']*\')|([^\s]*))"
            ).findall(line):
            if absolute and attr[1]!='=':
                print "** ABSOLUTE VALUES ONLY ** "+\
                      "No relative values are allowed for parameters"
                return (None,None)
            val=[]
            for tmpval in re.compile(
                "((\"[^\"]*\")|(\'[^\']*\')|([^,]+))\s*,*\s*"
                ).findall(attr[2]):
                tval=tmpval[0]
                tval=re.sub("\"([^\"]*)\"","\g<1>",tval)
                tval=re.sub("\'([^\']*)\'","\g<1>",tval)
                val.append(tval)
            field=self.getLDAPname(attr[0])
            if field:
                if attr[1]=="+=":
                    dict[field+"+"]=val
                elif attr[1]=="-=":
                    dict[field+"-"]=val
                else:
                    dict[field]=val
            else:
                print ("\n**UNKNOWN ATTRIBUTES** "+
                       "Unknown attribute '%s' specified.\n" % attr[0]+
                       "You may only specify the following "+
                       "attributes for a %s:\n" % self.name)
                for f in self.fields:
                    f.help()
                print
                return (None,None)
        return (dn,dict)

    def delete(self,dn):
        try:
            for subdir in clump.db.db.getFolder(dn):
                clump.db.db.delObject(subdir[0])
            clump.db.db.delObject(dn)
        except TypeError:
            print ("** NO SUCH %s ** No %s '%s' exists." %
                   (string.upper(self.name),self.name,attr['dn']))
            return
        except ldap.NO_SUCH_OBJECT:
            print ("** NO SUCH %s ** No %s '%s' exists." %
                   (string.upper(self.name),self.name,attr['dn']))
            return
        except ldap.NOT_ALLOWED_ON_NONLEAF:
            print ("** %s NOT EMPTY ** %s '%s' is not empty." %
                   (string.upper(self.name),string.capitalize(self.name),
                    attr['dn']))
            return
        log.write("Deleted %s %s" % (self.name,dn))
        return

    def set(self,line):
        dn,attr=self.parseLine(line)
        if not dn:
            return
        if not attr.has_key(self.namefield):
            name=re.compile("%s=([^,]+)" % self.namefield).match(dn).group(1)
        else:
            name=attr[self.namefield]
        # Check to see if there is an onSet function which should be called
        if self.__class__.__dict__.has_key("onSet"):
            obj=clump.db.db.getObject(dn)
            data={}
            for key in obj[1].keys():
                data[key]=obj[1][key]
            # Loop over all the fields from the command line and
            # merge them into the LDAP data entry correctly
            for key in attr.keys():
                # Check to see if we are adding a new entry to an
                # existing field
                if re.compile("\+$").search(key):
                    if data.has_key(key[:-1]):
                        data[key[:-1]]=data[key[:-1]]+attr[key]
                    else:
                        data[key[:-1]]=attr[key]
                # Check to see if we are removing an existing entry from
                # an existing field
                elif re.compile("\-$").search(key):
                    if not data.has_key(key[:-1]):
                        print ("** NO SUCH FIELD ** The object %s "+
                               "has no field '%s'") % (dn,key[:-1])
                        return
                    for item in attr[key]:
                        if not item in data[key[:-1]]:
                            print ("** NO SUCH ENTRY ** The object %s "+
                                   "has no entry '%s' in field '%s'") % (
                                dn,item,key[:-1])
                            return
                        else:
                            del data[key[:-1]][data[key[:-1]].index(item)]
                    data[key]=attr[key] # Add original field to cause delete
                # Finally just set the data field if its neither appending
                # nor removing entries
                else:
                    data[key]=attr[key]
            self.onSet(dn,data)
            mods={}
            for key in data.keys():
                if (not obj[1].has_key(key)) or data[key]!=obj[1][key]:
                    mods[key]=data[key]
            if mods:
                try:
                    clump.db.db.modObject((dn,mods))
		except ldap.NO_SUCH_OBJECT:
		    print ("** NO SUCH %s ** No %s '%s' exists." %
		           (string.upper(self.name),self.name,name))
		    return
                except ldap.NO_SUCH_ATTRIBUTE:
                    print ("** NO SUCH ATTRIBUTE ** "+
                           "Unable to delete attribute value.")
                    return
                except ldap.OBJECT_CLASS_VIOLATION:
                    print ("** REQUIRED ATTRIBUTE ** "+
                           "Attempt to delete a required attribute value.")
                    return
                except:
                    print "** FATAL ERROR ** Failed to modify object after"+\
                          " first pass"
                    return
        log.write("Modified %s %s" % (self.name,dn))
        return

    def create(self,line):
        # Call the method to extract the DN and attribute dictionary
        # with the create flag set
        dn,attr=self.parseLine(line,create=1)
        if not dn:   # If no DN was returned then quite creating now
            return 0
        return self.createNew(dn,attr)

    def createNew(self,dn,attr):
        # Get the defaults for the object
        defaults=getDefaults(self.name)
        # Loop over all the defaults and add them to the dictionary
        # is there isn't already a value specified
        for key in defaults.keys():
            if not attr.has_key(key):
                attr[key]=defaults[key]
        # Now loop over all the fields and make sure that there are
        # no missing creation attributes
        for f in self.fields:
            if f.create:
                if not attr.has_key(f.ldapname):
                    print (
                        "\n**MISSING ATTRIBUTES** "+
                        "Insufficient attributes to perform operation.\n"+
                        "You must specify the following attributes:\n")
                    for f2 in self.fields:
                        if f2.create:
                            f2.help()
                    print
                    return 0
        attr['objectClass']=self.objclass # Initialize the object classes
        # Set the calue of the field used to name the object using the
        # valued contained in the DN
        if not attr.has_key(self.namefield):
            attr[self.namefield]=[re.compile("[^=]+=([^,]+)").\
                                  match(dn).group(1)]
        # Look for an 'onSet' method and call it of there is one
        if self.__class__.__dict__.has_key("onSet"):
            self.onSet(dn,attr)
        # At last we attempt to create the object and catch any exceptions
        # which this may generate
        try:
            clump.db.db.addObject((dn,attr))
        except ldap.ALREADY_EXISTS:
            print("** ERROR ** %s '%s' already exists!" %
                  (string.capitalize(self.name),attr[self.namefield][0]))
            return 0
        # Finally we now loop and create all the subdirs for objects
        for subdir in self.subdirs:
            clump.db.db.addObject(("ou=%s," % subdir[0]+dn,
                             {'ou':subdir[0],'objectType':subdir[1],
                              'objectClass':['top','organizationalUnit',
                                             'clumpFolder']}))
        log.write("Created %s %s" % (self.name,dn))
        return -1

    def show(self,obj):
        print "\nFIELDS:"
        for field in self.fields:
            if field.show:
                try:
                    print "  %-10s : %s" % (field.name,string.join(
                        obj[1][field.ldapname],","))
                except KeyError:
                    print "  %-10s : <EMPTY>" % field.name
        print "\nCONTENTS:"
        names=[]
        for o in clump.db.db.getFolder(obj[0]):
            name=string.split(string.split(o[0],",")[0],"=")[1]
            if 'clumpFolder' in o[1]['objectClass']:
                name=name+">"
            names.append(name)
        names.sort()
        count=0
        text=""
        for name in names:
            text=text+" %-19s" % name
            count=count+1
            if (count % 4)==0:
                text=text+"\n"
        text=text+"\n"
        print text
        return
